/******************************************************************************
 * @file usart_driver.h
 * @brief USART/UART ISO 7816 buffered driver
 * @author Energy Micro AS
 * @version 1.00
 ******************************************************************************
 * @section License
 * <b>(C) Copyright 2013 Energy Micro AS, http://www.energymicro.com</b>
 *******************************************************************************
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 * 4. The source and compiled code may only be used on Energy Micro "EFM32"
 *    microcontrollers and "EFR4" radios.
 *
 * DISCLAIMER OF WARRANTY/LIMITATION OF REMEDIES: Energy Micro AS has no
 * obligation to support this Software. Energy Micro AS is providing the
 * Software "AS IS", with no express or implied warranties of any kind,
 * including, but not limited to, any implied warranties of merchantability
 * or fitness for any particular purpose or warranties against infringement
 * of any proprietary rights of a third party.
 *
 * Energy Micro AS will not be liable for any consequential, incidental, or
 * special damages, or any other relief, or for any claim by any third party,
 * arising from your use of this Software.
 *
 *****************************************************************************/
#ifndef __USART_DRIVER_H
#define __USART_DRIVER_H

typedef enum
{
  UART_STATUS_SUCCESS,
  UART_STATUS_NOT_ENOUGH_BYTES_RECEIVED,
  UART_STATUS_FRAMING_OR_PARITY_ERROR,
  UART_STATUS_USAGE_ERROR,
}UART_STATUS;

#define MAX_ERROR_COUNT (100)

/******************************************************************************
* @brief  usartSetup function
*
******************************************************************************/
void usartSetup(void);

/***************************************************************************//**
 * @brief
 *   Block/Unblock receiver
 *
 * @param[in] block
 *   true to block, false to unblock.
 ******************************************************************************/
void usartBlockReceiver(bool block);

/***************************************************************************//**
 * @brief
 *   Flush uart buffer, Always use this when switching between tx/rx operation.
 *
 ******************************************************************************/
void usartFlushBuffer(void);

/***************************************************************************//**
 * @brief
 *   Accept/not accept data on uart. Always use this to not accept data
 *   before sending data with uart. This makes sure TX bytes don't end up in rx
 *   queue.  
 *
 * @param[in] accept
 *   true to accept data and put in RX buffer, false to not accept data.
 ******************************************************************************/
void usartAcceptRX(bool accept);

/******************************************************************************
 * @brief  usartGetChar function
 *
 *  This function will return UART_STATUS based on what happens. 
 *
 * @param[out] receivedChar
 *   Pointer to unsigned char to put received data in. 
 *
 * @return
 *   UART_STATUS status enum.
 *****************************************************************************/
uint8_t usartGetChar(uint8_t * receivedChar);

/******************************************************************************
 * @brief  usartPutChar function
 *
 * This function will send single char return UART_STATUS based on what happens. 
 *
 * @param[in] ch
 *   unsigned char to transmit. 
 *
 * @return
 *   UART_STATUS status enum.
 *****************************************************************************/  
 uint8_t usartPutChar(uint8_t ch);

/******************************************************************************
 * @brief  usartPutData function
 *
 * This function will send char array and return UART_STATUS based on what happens. 
 *
 * @param[in] dataPtr
 *  pointer to unsigned char buffer to transmit. 
 *
 * @param[in] dataLen
 *  length of buffer to transmit.
 *
 * @return
 *   UART_STATUS status enum.
 *****************************************************************************/
uint8_t usartPutData(uint8_t * dataPtr, uint32_t dataLen);

/******************************************************************************
 * @brief  usartGetData function
 *
 * This function will try to receive char array and
 * return UART_STATUS based on what happens. 
 *
 * @param[in] dataPtr
 *  pointer to unsigned char buffer to receive to. 
 *
 * @param[in] dataLen
 *  length of buffer to receive, if 0, receive whatever is in receive buffer.
 *
 * @param[out] bytesReceived
 *  pointer to uint32_t, indicating how many bytes was actually received.
 *
 * @return
 *   UART_STATUS status enum.
 *****************************************************************************/
uint8_t usartGetData(uint8_t * dataPtr, uint32_t dataLen, uint32_t* bytesReceived);

/***************************************************************************//**
 * @brief Set up Clock Management Unit
 ******************************************************************************/
void cmuSetup(void);

#endif /* __USART_DRIVER_H */
